
import 'package:coders/ast/declarations.dart';
import 'package:coders/scanner/scanner.dart';
import 'package:coders/scanner/token.dart';
import 'parser_option.dart';
import 'package:coders/exception/exceptions.dart';
import 'package:coders/ast/ast_node.dart';
import 'package:coders/ast/statement.dart';

class Parser {
  final String source;

  final ParserOption option;

  Scanner scanner;

  /// current token.
  Token curToken;

  /// nextToken;
  Token nextToken;

  Parser(this.source, {this.option}) {
    scanner = Scanner(source);
  }

  void parse() {
    while (scanner.eof == false) {
      moveToken();

      parseDeclaration();

      if (nextToken.type == TokenType.Eof) {
        break;
      }
    }
  }

  void parseDeclaration() {
    if (nextToken.type == TokenType.Keyword) {
      final value = nextToken.value;

      if (value == 'struct') {
        parseStructDeclaration();
      } else if (value == 'type') {
        parseTypeDeclaration();
      } else if (value == 'table') {
        parseTableDeclaration();
      } else if (value == 'func') {
        parseFuncDeclaration();
      } else {
        assert(false);
      }
    }
  }

  Statement parseFuncDeclaration() {
    final decl = FuncDeclaration();

    decl.func = parseKeyword('func');

    decl.name = parseIdentifier();

    decl.lbraket = parsePunctuator('(');

    while (isNextPunctuator(')') == false) {
      decl.parameters.add(parseParameter());
      if (isNextPunctuator(',')) {
        parsePunctuator(',');
      }
    }

    decl.rbraket = parsePunctuator(')');

    decl.returnType = parseDataType();

    decl.block = parseBlockStatement();

    return decl;
  }

  BlockStatement parseBlockStatement() {
    var block = BlockStatement();

    block.lbrace = parsePunctuator('{');

    while(isNextPunctuator('}') == false) {
      block.stmt.add(parseStatement());
    }

    block.rbrace = parsePunctuator('}');

    return block;
  }

  Node parseStatement() {

    Node expr;

    switch(nextToken.type) {
      case TokenType.BooleanLiteral:
      case TokenType.NullLiteral:
      case TokenType.NumericLiteral:
      case TokenType.StringLiteral:
        expr = parseExpression();
        break;
      case TokenType.Punctuator:
        final value = nextToken.value;
        if(value == '{') {
          expr = parseBlockStatement();
        } else if(value == '(') {
          expr = parseExpression();
        } else if(value == ';') {
          expr = parseEmptyStatement();
        } else {
          expr = parseExpression();
        }
        break;
      case TokenType.Identifier:
        final value = nextToken.value;
        if(isSqlKeyword(value)) {
          expr = parseSqlLiteral();
        } else {
          expr = parseExpression();
        }
        break;
      case TokenType.Keyword:
        final value = nextToken.value;
        switch(value) {
          case 'break':
            expr = parseBreakStatement();
            break;
          case 'continue':
            expr = parseContinueStatement();
            break;
          case 'do':
            expr = parseDoWhileStatement();
            break;
          case 'for':
            expr = parseForStatement();
            break;
          case 'if':
            expr = parseIfStatement();
            break;
          case 'return':
            expr = parseReturnStatement();
            break;
          case 'switch':
            expr = parseSwitchStatement();
            break;
          case 'var':
            expr = parseVariableStatement();
            break;
          case 'while':
            expr = parseWhileStatement();
            break;
          case 'using':
            expr = parseUsingStatement();
            break;
          case 'commit':
            expr = parseCommitStatement();
            break;
          case 'rollback':
            expr = parseRollbackStatement();
            break;
          case 'call':
            expr = parseCallStatement();
            break;
          case 'fetch':
            expr = parseFetchStatement();
            break;
          default:
            throw UnexpectedToken(nextToken);
            break;
        }
        break;
      default:
        throw UnexpectedToken(nextToken);
        break;

    }

    return expr;
  }

  Expression parseExpression() {
    final expr = Expression();

    /// parse left side expression
    expr.left = parseLeftExpression();

    /// if next token is not assignment operator then
    /// return only left side expression.
    if(isNextAssignOperator() == false) {
      return expr;
    }

    expr.assignOperator = parseAssignOperator();

    expr.right = parseExpression();

    return expr;
  }

  UnaryOperator parseUnaryOperator() {
    final expr = UnaryOperator();
    final token = expectUnaryOperator();
    expr.token = token;
    return expr;
  }

  AssignOperator parseAssignOperator() {
    final expr = AssignOperator();
    final token = expectAssignOperator();
    expr.token = token;
    return expr;
  }

  UpdateOperator parseUpdateOperator() {
    final expr = UpdateOperator();
    final token = expectUpdateOperator();
    expr.token = token;
    return expr;
  }

  Node parseLeftExpression() {

    switch(nextToken.type)
    {
      case TokenType.StringLiteral:
        return parseValue();
      case TokenType.Identifier:
        return parseIdentifier();
      case TokenType.Punctuator:
        {
          final value = nextToken.value;
          if(value == '++' || value == '--') {
            return parseUpdateExpression();
          } else if(value == '+' || value == '-') {
            return parseUnaryExpression();
          } else {
            throw UnexpectedToken(nextToken);
          }
        }
        break;
    }

    throw UnexpectedToken(nextToken);
  }

  UnaryExpression parseUnaryExpression() {
    final expr = UnaryExpression();
    expr.operator = parseUnaryOperator();
    expr.name = parseIdentifierOrKeyword();
    return expr;
  }

  UpdateExpression parseUpdateExpression() {
    final expr = UpdateExpression();
    expr.operator = parseUpdateOperator();
    expr.name = parseIdentifierOrKeyword();
    return expr;
  }

  Statement parseFetchStatement() {
    throw UnimplementedError();
  }

  Statement parseCallStatement() {
    throw UnimplementedError();
  }

  Statement parseRollbackStatement() {
    throw UnimplementedError();
  }

  Statement parseUsingStatement() {
    throw UnimplementedError();
  }

  Statement parseCommitStatement() {
    throw UnimplementedError();
  }

  Statement parseWhileStatement() {
    throw UnimplementedError();
  }

  Statement parseVariableStatement() {
    var decl = VarDeclaration();

    decl.varKeyword = parseKeyword('var');

    decl.name = parseIdentifierOrKeyword();

    decl.dataType = parseDataType();
    if(isNextPunctuator('=')) {
      decl.equal = parsePunctuator('=');
      decl.defaultValue = parseValue();
    }

    decl.semicolon = parsePunctuator(';');

    return decl;
  }

  Statement parseSwitchStatement() {
    throw UnimplementedError();
  }

  Statement parseReturnStatement() {
    throw UnimplementedError();
  }

  Statement parseIfStatement() {
    throw UnimplementedError();
  }

  Statement parseForStatement() {
    throw UnimplementedError();
  }

  Statement parseDoWhileStatement() {
    throw UnimplementedError();
  }

  Statement parseContinueStatement() {
    throw UnimplementedError();
  }

  Statement parseBreakStatement() {
    throw UnimplementedError();
  }

  Statement parseSqlLiteral() {
    throw UnimplementedError();
  }

  Node parseEmptyStatement() {
    return parsePunctuator(';');
  }


  Statement parseVarDeclaration() {

    var decl = VarDeclaration();

    decl.varKeyword = parseKeyword('var');

    decl.name = parseIdentifier();

    decl.dataType = parseDataType();

    if(isNextPunctuator('=')) {

      decl.equal = parsePunctuator('=');

      decl.defaultValue = parseValue();
    }

    decl.semicolon = parsePunctuator(';');

    return decl;
  }

  Statement parseParameter() {
    final decl = Parameter();

    if (isNextKeyword('in')) {
      decl.inout = parseKeyword('in');
    }

    if (isNextKeyword('out')) {
      decl.inout = parseKeyword('out');
    }

    if (isNextKeyword('inout')) {
      decl.inout = parseKeyword('inout');
    }

    decl.name = parseIdentifier();

    decl.dataType = parseDataType();

    return decl;
  }

  Statement parseDataType() {
    final expr = DataType();

    expr.name = parseIdentifierOrKeyword();

    final value = curToken.value;

    if (curToken.type == TokenType.Keyword) {
      if (value == 'int8' ||
          value == 'int16' ||
          value == 'int32' ||
          value == 'float' ||
          value == 'double' ||
          value == 'bigint' ||
          value == 'integer' ||
          value == 'string') {
        // do nothing.
      } else if (value == 'varchar') {
        expr.lbraket = parsePunctuator('(');
        expr.length = parseNumeric();
        expr.rbraket = parsePunctuator(')');
      } else if (value == 'char') {
        if (isNextPunctuator('(')) {
          expr.lbraket = parsePunctuator('(');
          expr.length = parseNumeric();
          expr.rbraket = parsePunctuator(')');
        }
      } else {
        throw UnexpectedToken(curToken);
      }
    } else {
      throw UnexpectedToken(curToken);
    }
    return expr;
  }

  Statement parseTableDeclaration() {
    final decl = TableDeclaration();

    decl.table = parseKeyword('table');

    decl.name = parseIdentifierOrKeyword();

    decl.lbrace = parsePunctuator('{');

    final members = <String>{};

    while (isNextPunctuator('}') == false) {
      final value = nextToken.value;
      final type = nextToken.type;

      if (type == TokenType.Keyword) {
        if (value == 'key') {
          parseTableKeyDeclaration();
        } else if (value == 'index') {
          parseTableIndexDeclaration();
        } else if (value == 'link') {
          parseTableLinkDeclaration();
        } else {
          decl.memberDecls.add(parseTableMemberDeclaration(members));
        }
      } else {
        decl.memberDecls.add(parseTableMemberDeclaration(members));
      }
    }

    decl.rbrace = parsePunctuator('}');

    return decl;
  }

  Statement parseTableLinkDeclaration() {
    throw UnimplementedError();
  }

  Statement parseTableKeyDeclaration() {
    throw UnimplementedError();
  }

  Statement parseTableIndexDeclaration() {
    throw UnimplementedError();
  }

  Statement parseTypeDeclaration() {
    final decl = TypeDeclaration();

    decl.type = parseKeyword('type');

    decl.name = parseIdentifier();

    decl.dataType = parseDataType();

    decl.semicolon = parsePunctuator(';');

    return decl;
  }

  Statement parseTableMemberDeclaration(Set<String> members) {
    final decl = TableMemberDeclaration();

    decl.name = parseIdentifierOrKeyword();

    if (members.contains(decl.name.value)) {
      throw DuplicateToken(decl.name.token);
    } else {
      members.add(decl.name.value);
    }

    decl.dataType = parseDataType();

    if (isNextPunctuator('=')) {
      decl.equal = parsePunctuator('=');
      decl.defaultValue = parseValue();
    }

    while (isNextPunctuator(';') == false) {
      if (isNextKeyword('nullable')) {
        decl.nullable = parseKeyword('nullable');
      } else if (isNextKeyword('auto')) {
        decl.auto = parseKeyword('auto');
      } else {
        throw UnexpectedToken(curToken);
      }
    }

    decl.semicolon = parsePunctuator(';');

    return decl;
  }

  Statement parseStructDeclaration() {
    final decl = StructDeclaration();

    decl.struct = parseKeyword('struct');

    decl.name = parseIdentifier();

    decl.lbrace = parsePunctuator('{');

    final members = <String>{};

    while (isNextPunctuator('}') == false) {
      decl.memberDecls.add(parseStructMemberDeclaration(members));
    }

    decl.rbrace = parsePunctuator('}');

    return decl;
  }

  StructMemberDeclaration parseStructMemberDeclaration(Set<String> members) {
    final decl = StructMemberDeclaration();

    decl.name = parseIdentifier();

    if (members.contains(decl.name.value)) {
      throw DuplicateToken(decl.name.token);
    } else {
      members.add(decl.name.value);
    }

    decl.dataType = parseDataType();

    decl.semicolon = parsePunctuator(';');

    return decl;
  }

  Numeric parseNumeric() {
    final expr = Numeric();
    final token = expectNumeric();
    expr.token = token;
    return expr;
  }

  Value parseValue() {
    final expr = Value();
    final token = expectValue();
    expr.token = token;
    return expr;
  }

  Punctuator parsePunctuator(String str) {
    final expr = Punctuator();
    final token = expectPunctuator(str);
    expr.token = token;
    return expr;
  }

  /// 키워드를 식별자로 강제로 수정한다.
  /// 키워드를 식별자로 사용하는 경우 호출한다.
  Identifier parseIdentifierOrKeyword() {
    final expr = Identifier();
    final token = expectIdentifierOrKeyward();
    expr.token = token;
    return expr;
  }

  Identifier parseIdentifier() {
    final expr = Identifier();
    final token = expectIdentifier();
    expr.token = token;
    return expr;
  }

  Keyword parseKeyword(String keyword) {
    final expr = Keyword();
    final token = expectKeyword(keyword);
    expr.token = token;
    return expr;
  }

  bool isNextAssignOperator() {
    if(nextToken.type != TokenType.Punctuator) {
      return false;
    }

    if(nextToken.value == '=') {
      return true;
    }

    return false;
  }

  bool isSqlKeyword(String value) {
    return value == 'select' ||
        value == 'update' ||
        value == 'delete' ||
        value == 'insert';
  }

  bool isNextDataType() {
    var value = nextToken.value;
    switch (value) {
      case 'char':
      case 'int8':
      case 'int16':
      case 'int32':
      case 'int64':
      case 'float':
      case 'double':
      case 'string':
      case 'varchar':
        return true;
    }
    return false;
  }

  bool isNextKeyword(String keyword) {
    if (nextToken.type == TokenType.Keyword && nextToken.value == keyword) {
      return true;
    } else {
      return false;
    }
  }

  bool isNextNumeric() {
    if (nextToken.type == TokenType.NumericLiteral) {
      return true;
    } else {
      return false;
    }
  }

  bool isNextString() {
    if (nextToken.type == TokenType.StringLiteral) {
      return true;
    } else {
      return false;
    }
  }

  /// check next token punctuator
  bool isNextPunctuator(String str) {
    if (nextToken.type == TokenType.Punctuator && nextToken.value == str) {
      return true;
    } else {
      return false;
    }
  }

  Token expectUnaryOperator() {
    moveToken();

    if(curToken.type != TokenType.Punctuator) {
      throw UnexpectedToken(curToken);
    }

    final value = curToken.value;

    if(value == '+' || value == '-') {
      return curToken;
    } else {
      throw UnexpectedToken(curToken);
    }
  }

  Token expectUpdateOperator() {
    moveToken();

    if(curToken.type != TokenType.Punctuator) {
      throw UnexpectedToken(curToken);
    }

    final value = curToken.value;

    if(value == '++' || value == '--') {
      return curToken;
    } else {
      throw UnexpectedToken(curToken);
    }
  }

  Token expectAssignOperator() {

    moveToken();

    if(curToken.type != TokenType.Punctuator) {
      throw UnexpectedToken(curToken);
    }

    final value = curToken.value;

    if(value == '=' || value == '+=' || value == '-=' || value == '*=' || value == '|=') {
      return curToken;
    } else {
      throw UnexpectedToken(curToken);
    }
  }

  Token expectValue() {
    moveToken();

    if (curToken.type != TokenType.NumericLiteral &&
        curToken.type != TokenType.StringLiteral) {
      throw UnexpectedToken(curToken);
    }
    return curToken;
  }

  Token expectNumeric() {
    moveToken();

    if (curToken.type != TokenType.NumericLiteral) {
      throw UnexpectedToken(curToken);
    }

    return curToken;
  }

  Token expectPunctuator(String str) {
    moveToken();

    if (curToken.type != TokenType.Punctuator || curToken.value != str) {
      throw UnexpectedToken(curToken);
    }
    return curToken;
  }

  Token expectIdentifier() {
    moveToken();

    if (curToken.type != TokenType.Identifier) {
      throw UnexpectedToken(curToken);
    }
    return curToken;
  }

  Token expectIdentifierOrKeyward() {
    moveToken();

    if (curToken.type == TokenType.Identifier ||
        curToken.type == TokenType.Keyword) {
      return curToken;
    }

    throw UnexpectedToken(curToken);
  }

  Token expectKeyword(String keyword) {
    moveToken();

    if (curToken.type != TokenType.Keyword) {
      throw UnexpectedToken(curToken);
    }

    if (keyword != null) {
      if (curToken.value != keyword) {
        throw UnexpectedToken(curToken);
      }
    }

    return curToken;
  }

  void moveToken() {
    curToken = nextToken;
    nextToken = scanner.scan();
  }
}
