import 'character.dart';
import 'char.dart';
import 'comment.dart';
import 'package:coders/exception/exceptions.dart';
import 'keyword.dart';
import 'location.dart';
import 'token.dart';

class Scanner {
  /// source
  final String source;

  /// length of source
  int length = 0;

  /// current line of source
  int curLine = 0;

  /// begin index of current line
  int begLineIndex = 0;

  /// begin index of token value
  /// initialized when begLocation function called
  int begIndex = 0;

  /// end index of token value
  /// initialized when endLocation function called.
  int endIndex = 0;

  /// current index
  int curIndex = 0;

  /// location of token
  var location = Location();

  /// comments
  var comments = <Comment>[];

  Scanner(this.source) {
    length = source.length;
  }

  /// calculate column of value
  int get column => curIndex - begLineIndex;

  /// calc value of token
  String get value {
    if (length <= curIndex) {
      return source.substring(begIndex);
    } else {
      return source.substring(begIndex, endIndex);
    }
  }

  bool get eof => curIndex >= length;

  int codeUnit(int index) {
    if (length <= index) {
      return 0;
    } else {
      return source.codeUnits[index];
    }
  }

  bool isKeyword(String id) {
    return Keyword.contains(id);
  }

  Token newToken(TokenType type, String value) {
    final token = Token(type, value, comments, location);
    comments = [];
    return token;
  }

  void snapBeginLocation() {
    begIndex = curIndex;
    location.begColumn = column;
    location.begLine = curLine;
  }

  void snapEndLocation({int offset}) {
    endIndex = curIndex;
    location.endLine = curLine;

    if (offset == null) {
      location.endColumn = column;
    } else {
      location.endColumn = column + offset;
    }
  }

  Token scan() {
    scanComments();

    if (eof) {
      return newToken(TokenType.Eof, '');
    }

    final cp = codeUnit(curIndex);

    if (Character.isIdentifierStart(cp)) {
      return scanIdentifier();
    }

    /// common ( ) ;
    if (cp == Char.LParan || cp == Char.RParan || cp == Char.Semicolon) {
      return scanPunctuator();
    }

    if (cp == Char.Quete || cp == Char.DoubleQuete) {
      return scanStringLiteral();
    }

    if (Character.isDecimalDigit(cp)) {
      return scanNumericLiteral();
    }

    if (cp == Char.Dot) {
      if (Character.isDecimalDigit(codeUnit(curIndex + 1))) {
        return scanNumericLiteral();
      }
    }

    return scanPunctuator();
  }

  Token scanStringLiteral() {
    snapBeginLocation();

    final quote = codeUnit(curIndex);

    ++curIndex;

    while (eof == false) {
      final cp = codeUnit(curIndex);

      ++curIndex;

      if (cp == quote) {
        snapEndLocation();

        return newToken(TokenType.StringLiteral, value);
      }

      if (cp == '\\'.codeUnits.first) {
        final cp2 = codeUnit(curIndex);

        if (cp2 == quote) {
          ++curIndex;
        }
      }
    }

    snapEndLocation();

    throw UnexpectedToken.location(location, value);
  }

  Token scanNumericLiteral() {
    snapBeginLocation();

    var cp = codeUnit(curIndex);

    ++curIndex;

    if (cp != '.'.codeUnits.first) {
      if (cp == Char.num_0) {
        var cp2 = codeUnit(curIndex);

        if (cp2 == Char.x || cp2 == Char.X) {
          return scanHexLiteral();
        }

        if (cp2 == Char.b || cp2 == Char.B) {
          return scanBinaryLiteral();
        }

        if (cp2 == Char.o || cp2 == Char.O) {
          return scanOctalLiteral();
        }
      }

      while (Character.isOctalDigit(codeUnit(curIndex))) {
        ++curIndex;
      }

      cp = codeUnit(curIndex);
    }

    if (cp == '.'.codeUnits.first) {
      while (Character.isDecimalDigit(codeUnit(curIndex))) {
        ++curIndex;
      }

      cp = codeUnit(curIndex);
    }

    if (cp == Char.e || cp == Char.E) {
      ++curIndex;

      cp = codeUnit(curIndex);

      if (cp == Char.Plus || cp == Char.Minus) {
        ++curIndex;
      }

      if (Character.isDecimalDigit(codeUnit(curIndex))) {
        while (Character.isDecimalDigit(codeUnit(curIndex))) {
          ++curIndex;
        }
      } else {
        snapEndLocation();

        throw UnexpectedToken.location(location, value);
      }
    }

    if (Character.isIdentifierStart(codeUnit(curIndex))) {
      ++curIndex;
      snapEndLocation();

      throw UnexpectedToken.location(location, value);
    }

    snapEndLocation(offset: -1);

    return newToken(TokenType.NumericLiteral, value);
  }

  Token scanOctalLiteral() {
    ++curIndex;

    while (Character.isOctalDigit(codeUnit(curIndex))) {
      ++curIndex;
    }

    if (Character.isIdentifierStart(codeUnit(curIndex))) {
      throw UnexpectedToken.location(location, value);
    }

    snapEndLocation(offset: -1);

    return newToken(TokenType.NumericLiteral, value);
  }

  Token scanBinaryLiteral() {
    ++curIndex;

    while (Character.isBinaryDigit(codeUnit(curIndex))) {
      ++curIndex;
    }

    if (Character.isIdentifierStart(codeUnit(curIndex))) {
      throw UnexpectedToken.location(location, value);
    }

    snapEndLocation(offset: -1);

    return newToken(TokenType.NumericLiteral, value);
  }

  Token scanHexLiteral() {
    ++curIndex;

    while (Character.isHexDigit(codeUnit(curIndex))) {
      ++curIndex;
    }

    if (Character.isIdentifierStart(codeUnit(curIndex))) {
      throw UnexpectedToken.location(location, value);
    }

    snapEndLocation(offset: -1);

    return newToken(TokenType.NumericLiteral, value);
  }

  void scanComments() {
    comments = [];

    while (eof == false) {
      var cp = codeUnit(curIndex);

      if (Character.isWhiteSpace(cp)) {
        ++curIndex;
      } else if (Character.isLineTerminator(cp)) {
        ++curIndex;
        if (cp == Char.CR && codeUnit(curIndex) == Char.LF) {
          ++curIndex;
        }

        ++curLine;

        begLineIndex = curIndex;
      } else if (cp == Char.Slash) {
        var cp = source.codeUnits[curIndex + 1];

        if (cp == Char.Slash) {
          snapBeginLocation();

          curIndex += 2;

          var comment = scanSingleLineComment();

          comments.add(comment);
        } else if (cp == Char.Star) {
          snapBeginLocation();

          curIndex += 2;

          var comment = scanMultiLineComment();

          comments.add(comment);
        } else {
          break;
        }
      } else {
        break;
      }
    }
  }

  Comment scanMultiLineComment() {
    while (eof == false) {
      final cp = codeUnit(curIndex);

      if (Character.isLineTerminator(cp)) {
        if (cp == Char.CR && source.codeUnits[curIndex + 1] == Char.LF) {
          ++curIndex;
        }

        ++curLine;

        ++curIndex;

        begLineIndex = curIndex;
      } else if (cp == Char.Star) {
        if (source.codeUnits[curIndex + 1] == Char.Slash) {
          curIndex += 2;

          snapEndLocation();

          return Comment(CommentType.Block, value, location);
        }
      } else {
        ++curIndex;
      }
    }

    snapEndLocation();

    return Comment(CommentType.Block, value, location);
  }

  Comment scanSingleLineComment() {
    Comment comment;

    while (eof == false) {
      var cp = codeUnit(curIndex);

      if (Character.isLineTerminator(cp)) {
        snapEndLocation();

        if (cp == Char.CR && source.codeUnits[curIndex + 1] == Char.LF) {
          ++curIndex;
        }

        ++curLine;

        begLineIndex = curIndex;

        comment = Comment(CommentType.Line, value, location);

        return comment;
      }

      ++curIndex;
    }

    snapEndLocation();

    return Comment(CommentType.Line, value, location);
  }

  Token scanPunctuator() {
    snapBeginLocation();

    var cp = codeUnit(curIndex);

    if (cp == Char.LParan || cp == Char.LBrace) {
      ++curIndex;
      snapEndLocation();
      return newToken(TokenType.Punctuator, value);
    }

    if (cp == Char.Dot) {
      ++curIndex;
      snapEndLocation();
      return newToken(TokenType.Punctuator, value);
    }

    if (cp == Char.RBrace) {
      ++curIndex;
      snapEndLocation();
      return newToken(TokenType.Punctuator, value);
    }

    if (cp == Char.RParan ||
        cp == Char.Semicolon ||
        cp == Char.Comma ||
        cp == Char.LBraket ||
        cp == Char.RBraket ||
        cp == Char.Colon ||
        cp == Char.Question ||
        cp == Char.Bitnot ||
        cp == Char.At) {
      ++curIndex;
      snapEndLocation();

      return newToken(TokenType.Punctuator, value);
    }

    final cp2 = codeUnit(curIndex + 1);

    if ((cp == Char.Equal && cp2 == Char.LessThan) // =<
        || (cp == Char.Equal && cp2 == Char.MoreThan) // =>
        || (cp == Char.Equal && cp2 == Char.Equal) // ==
        || (cp == Char.Plus && cp2 == Char.Equal) // +=
        || (cp == Char.Minus && cp2 == Char.Equal) // -=
        || (cp == Char.Multiply && cp2 == Char.Equal) // *=
        || (cp == Char.Divide && cp2 == Char.Equal) // /=
        || (cp == Char.Remainder && cp2 == Char.Equal) // %=
        || (cp == Char.Plus && cp2 == Char.Plus) // ++
        || (cp == Char.Minus && cp2 == Char.Minus) // --
        || (cp == Char.Bitor && cp2 == Char.Bitor) // ||
        || (cp == Char.Bitand && cp2 == Char.Bitand) // &&
    ) {
      curIndex += 2;
      snapEndLocation();
      return newToken(TokenType.Punctuator, value);
    }

    if(cp == Char.Equal) {
      ++curIndex;
      snapEndLocation();
      return newToken(TokenType.Punctuator, value);
    }

    ++curIndex;

    snapEndLocation();

    throw UnexpectedToken.location(location, value);
  }

  Token scanIdentifier() {
    snapBeginLocation();

    ++curIndex;

    while (Character.isIdentifierPart(codeUnit(curIndex))) {
      ++curIndex;
    }

    snapEndLocation(offset: -1);

    final id = value;

    TokenType type;
    if (id.length == 1) {
      type = TokenType.Identifier;
    } else if (isKeyword(id)) {
      type = TokenType.Keyword;
    } else if (id == 'null') {
      type = TokenType.NullLiteral;
    } else if (id == 'true' || id == 'false') {
      type = TokenType.BooleanLiteral;
    } else {
      type = TokenType.Identifier;
    }

    return newToken(type, id);
  }
}
